''********************************************
''*  PropTerm S100 I/O Driver               *
''********************************************
''
''
''This object maintains two fifos in global memory to allow for other cogs
''to read/write data to/from the USB interface.
''  
con
  c_RxFifoSize  = 128           'bytes in the rx fifo, must be power of 2 that is >= 8
  c_TxFifoSize  =  16           'bytes in the tx fifo, must be power of 2 that is >= 2
  
VAR

  long  cog                     'cog flag/id
  'first 4 longs are fifo variables
  long  rx_head                 '
  long  rx_tail
  long  tx_head
  long  tx_tail
  'next the bit masks for our I/O pins
  long  buffer_ptr
  'rx/tx fifo data buffers
  byte  rx_buffer[c_RxFifoSize] 'receive buffer
  byte  tx_buffer[c_TxFifoSize] 'transmit buffer

PUB Start : okay
''Start serial driver - starts a cog
''returns false if no cog available
  Stop
  'initalize RX/TX fifos to empty state 
  longfill(@rx_head, 0, 4)
  'Initalize RX FIFO data start address
  buffer_ptr := @rx_buffer
    
'Start COG
  okay := cog := cognew(@entry, @rx_head) + 1

PRI CalcBitMask(Bit) : Mask
'Calculate a long mask with the bit number "Bit" set to one,  all other bits zero
  Mask := 1 << Bit
  
PUB Stop
''Stop S100 bus driver if cog variable <> 0 - frees a cog
  if cog
    cogstop(cog~ - 1)

PUB ChangeDataFormat(aBaudRate, aDataBits, aParity)
''Change RS232 data rate and/or format, RX and TX
''Has no effect, stub to making changing interfaces (RS232/S100 bus) simple
  aDataBits:= aParity
  
PUB GetRxHeadAdr : Address
''Return memory address of receive fifo head variable
  Address:= @rx_head
  
PUB GetRxTailAdr : Address
''Return memory address of receive fifo tail variable
  Address:= @rx_Tail
  
PUB GetRxBufAdr : Address
''Return memory address of receive fifo data byte array
  Address:= @rx_Buffer
  
PUB GetTxHeadAdr : Address
''Return memory address of transmit fifo head variable
  Address:= @tx_head
  
PUB GetTxTailAdr : Address
''Return memory address of transmit fifo tail variable
  Address:= @tx_Tail
  
PUB GetTxBufAdr : Address
''Return memory address of transmit fifo data byte array
  Address:= @tx_Buffer

DAT

'**********************************************************
'* Assembly language code  S100 bus interface             *
'**********************************************************
'
                        org
                        
entry                   mov     t1, par                 'get structure address
                        add     t1, #4 << 2             'skip past heads and tails
                        rdlong  RxBuffAdr, t1

                        mov     TxBuffAdr, RxBuffAdr    'rx buffer is first
                        add     TxBuffAdr, #c_RxFifoSize'tx starts c_RxFifoSize bytes later
                        'Setup fifo addresses and read shadow values.
                        'Read current RxHead to shadow RXHead
                        mov     RxHeadAdr, par          'read address of rx head
                        rdlong  SdwRxHead, RxHeadAdr    'SdwRxHead = rx head
                        'Calc RxTail address
                        mov     RxTailAdr, RxHeadAdr
                        add     RxTailAdr, #4
                        'Calc TxHead address 
                        mov     TxHeadAdr, RxTailAdr
                        add     TxHeadAdr, #4 
                        'Calc TxTail address 
                        mov     TxTailAdr, TxHeadAdr
                        add     TxTailAdr, #4 
                        rdlong  SdwTxTail, TxTailAdr    'SdwTxTail = tx tail
'
                        cmp     t1, t1            wz    'Set cog Zero flag to 1
                        mov     dira,IOMask
                        mov     outa,SetOutOne
                        mov     kbstat,#0               ' keyboard_status := 0
                        mov     tdL,#0
                        mov     tdR,#0
                        
MainLoop                test    p11mask, ina   wz       '  if ina[11] == %1
              if_z      jmp     #DoSend                 'if Rx ready is not = 0 no data to read
                        'We have data at the S100 bus ready to read, see if local RX FIFO has room.
                        'Check for, fifo full, head + 1 = tail
                        mov     t3, SdwRxHead           'T3 = rx head
                        add     t3, #1                  'T3 = head + 1, still need to wrap
                        cmp     t3, #c_RxFifoSize wz    'T3 = head +1 mod c_RxFifoSize, in case head wrapped
              if_z      mov     t3, #0                        
                        rdlong  t2, RxTailAdr           'T2 = rx tail
                        cmp     t2, t3            wz    'if head + 1 = tail, fifo full
              if_z      jmp     #DoSend                 'if Rx FIFO full do not read S100 bus, try a write
                        'S100 bus has data and we have room so read S100 bus and place data in local RX FIFO
                        'First read S100 bus data
                        cmp     t1, t1            wz     'Set cog Zero flag to 1
                        muxnz   outa,p25mask             ' outa[25] := 0 
                        muxz    outa, #%100000000        'outa[8] := 1
                        muxnz   dira,#$FF                'dira[7..0]~
                        mov     t1, ina                 'Read data bits
                        muxz    outa,p25mask             ' outa[25] := 1                        
                        muxz    dira,#$FF           'dira[7..0]~~
                        muxnz   outa, #%100000000        'outa[8] := 0
                        
                        mov     t2, SdwRxHead           'T2 = current rxHead value
                        add     t2, RxBuffAdr           'Add rxBuffer global memory address
                        wrbyte  t1, t2                  'Save the new data to the fifo
                        mov     SdwRxHead, t3           'T3 = RXxHead + 1, Point head to next location
                        wrlong  SdwRxHead, RxHeadAdr    'Update the rx FIFO Head in global memory

                        cmp     t1, t1            wz          'Set cog Zero flag to 1
                        muxz    dira,#$FF                     'dira[7..0]~~   
                        mov     t2,outa
                        andn    t2,#$FF
                        and     t1,#$FF
                        or      t2,t1
                        mov     outa,t2
                        muxnz   outa,p26mask               ' strobe p27
                        nop
                        muxz    outa,p26mask                                          

                        cmp     t1,#7              wz
                 if_z   jmp     #Bell       

DoSend                        
                        cmp     kbstat,#0       wz
                if_nz   jmp     #ws100rd                  ' if keyboard_start <>0 then wait
 'End of receive code now for transmit code.
                        'First see if TX FIFO is full
                        rdlong  t2,TxHeadAdr            'T2 = txHead
                        cmp     t2, SdwTxTail   wz      'see if Head = Tail, indicates nothing to send
              if_z      jmp     #MainLoop
                        'Local TX FIFO has a byte to send and USB is ready to accept it.
                        'Read byte from local tx fifo first
                        'There is at least one byte to read
                        mov     kbstat,#$FF
                        mov     tdL,tdR                 'shift left / right display digits
                        mov     t3, SdwTxTail           'Read the tx tail from our cache
                        add     t3, TxBuffAdr           'Calc address of next fifo byte to read
                        rdbyte  t1, t3                  'T1 = next byte to send
                        mov     tdR,t1
                        'Data in T1, now inc tx tail
                        add     SdwTxTail, #1                 'point to next fifo location
                        cmp     SdwTxTail, #c_TxFifoSize wz   'account for possible wrap
              if_z      mov     SdwTxTail, #0                 'point to first fifo buffer location                        
                        wrlong  SdwTxTail, TxTailAdr          'save tx tail to global memory
'                        cmp     t1, t1            wz          'Set cog Zero flag to 1
'                        muxz    dira,#$FF                     'dira[7..0]~~   
'                        mov     t2,outa
'                        andn    t2,#$FF
'                        or      t2,tdL
'                        mov     outa,t2
'                        cmp     t1, t1            wz       'Set cog Zero flag to 1
'                        muxnz   outa,p26mask               'strobe p26
'                        nop
'                        muxz    outa,p26mask
'                        cmp     t1, t1            wz        'Set cog Zero flag to 1
'                        muxz    dira,#$FF
'                        mov     t2,outa
'                        andn    t2,#$FF
'                        or      t2,tdR
'                        mov     outa,t2
'                        cmp     t1, t1            wz       ' Set cog Zero flag to 1
'                        muxnz   outa,p27mask               ' strobe p27
'                        nop
'                        muxz    outa,p27mask                        

                        cmp     t1, t1            wz          'Set cog Zero flag to 1
                        mov     t2,outa
                        andn    t2,#$FF
                        or      t2,t1
                        mov     outa,t2
                        muxnz   outa,p27mask               ' strobe p27
                        nop
                        muxz    outa,p27mask                        

                        muxnz   outa,p24mask               ' strobe p24
                        nop
                        muxz    outa,p24mask

                        muxz    outa,p10mask
                        
                        muxnz   outa,p12mask
                        mov     t1,p12delay
p12wait                 nop
                        djnz    t1,#p12wait
                        cmp     t1,t1                                                                   
                        muxz   outa,p12mask
                        
                        jmp     #MainLoop

ws100rd                 test    p9mask, ina   wz       '  if ina[9] == %1                                                   
                 if_nz  jmp     #MainLoop              
                        muxnz    outa,p10mask          ' outa[10] := 0
                        mov     kbstat,#0
                        jmp     #MainLoop
                        
Bell                    mov     t1,belldelay
                        muxnz   outa,p13mask                          
p12b                    nop
                        djnz    t1,#p12b                        
                        muxz    outa,p13mask
                        jmp     #DoSend  
                           
'
D0                      long    (1 << 9)
truc2                   long    (1<<  26)
trucFF                  long    ($FFFFFFFF)
DataMask                long    ($000000FF)
DataShift               long     16
IOMask                  long     %00001111000000000011010111111111
SetOutOne               long     %00001111000000000011000000000000
p9mask                  long     %1000000000
p10mask                 long     %10000000000
p11mask                 long     %100000000000
p12mask                 long     %1000000000000
p13mask                 long     %10000000000000
p24mask                 long     %1000000000000000000000000
p25mask                 long     %10000000000000000000000000
p26mask                 long     %100000000000000000000000000
p27mask                 long     %1000000000000000000000000000
kbstat                  long     0
p12delay                long     2000 '100000
belldelay               long    600000
'
' Uninitialized data
'
SdwRxHead               res     1         'Only this cog updates rxHead so we keep a shadow copy for speed
SdwTxTail               res     1         'Only this cog updates txTail so we keep a shadow copy for speed
'
'Addresses to speed reading of global data
RxHeadAdr               res     1
RxTailAdr               res     1
TxHeadAdr               res     1
TxTailAdr               res     1
'
'Temp use variables
t1                      res     1
t2                      res     1
t3                      res     1
tdL                     res     1         ' save area left digit
tdR                     res     1         ' save area right digit
'
'I/O masks for pins, these must stay in same order as in VAR section 
RxBuffAdr               res     1
TxBuffAdr               res     1